from enum import Enum


class PressureUnit(str, Enum):
    PASCAL = "Pa"
    KILOPASCAL = "kPa"
    MEGAPASCAL = "mPa"
    BAR = "bar"


class TemperatureUnit(str, Enum):
    CELSIUS = "C"
    FAHRENHEIT = "F"
    KELVIN = "K"


class UnitSystem(str, Enum):
    SI = "SI"
    SI_WITH_CELISUS = "SIC"
    IMPERIAL = "IS"



from typing import Optional

import yaml
from pydantic import Field
from pydantic_settings import BaseSettings

from mixprops.units import PressureUnit, UnitSystem


class Settings(BaseSettings):
    units: UnitSystem = Field(default=UnitSystem.SI_WITH_CELISUS)
    pressure_unit: Optional[PressureUnit] = Field(default=PressureUnit.PASCAL)


def get_settings() -> Settings:
    try:
        with open("settings.yaml") as file:
            settings_yaml = yaml.safe_load(file)
            settings = Settings.model_validate(settings_yaml)
    except IOError:
        settings = Settings()
    return settings



import os
from typing import List

import pandas as pd

from mixprops.species import Species

MODULE_PATH = os.path.dirname(os.path.abspath(__file__))
DATA_PATH = os.path.join(MODULE_PATH, "data")


class SpeciesNotFoundException(Exception):
    def __init__(self, name: str):
        self.message = f"Species {name} not found"
        super().__init__(self.message)


class Database:
    def __init__(self, species: List[Species]):
        self.species = species

    @classmethod
    def from_file(cls, file: str):
        df = pd.read_csv(DATA_PATH + f"/{file}", index_col=0)
        species_list = df.to_dict(orient="records")
        return cls(species=[Species(**item) for item in species_list])

    def find_species(self, name: str) -> Species:
        species = [
            item for item in self.species if item.species == name or item.name == name
        ]
        if not species:
            raise SpeciesNotFoundException(name=name)
        return species[0]



from typing import List, Tuple

import numpy as np
import pandas as pd
from pydantic import BaseModel, ConfigDict, Field, computed_field

from mixprops.constants import R_U
from mixprops.database import Database
from mixprops.species import MixtureSpecies
from mixprops.state import State


def is_valid_composition(number: float, total: float) -> bool:
    return 0 <= number <= total


class Mixture(BaseModel):
    species: List[Tuple[str, float]] = Field(repr=False, exclude=True)
    state: State = Field(repr=False, exclude=True)
    df: pd.DataFrame = Field(repr=False, exclude=True)
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(
        self,
        species: List[Tuple[str, float]],
        state: State,
    ):
        total = 1.0
        composition = [spec[1] for spec in species]

        if not all(is_valid_composition(value, total) for value in composition):
            raise ValueError("Invalid species composition, expecting fraction format")

        if abs(sum(composition) - total) > 1e-6:
            raise ValueError(f"Invalid species composition, they should sum to {total}")

        db: Database = Database.from_file(
            file="species_T_200_1000.csv"
            if state.temperature <= 1000
            else "species_T_1000_6000.csv"
        )
        relevant_species: List[MixtureSpecies] = []

        for spec in species:
            relevant_spec = db.find_species(name=spec[0])
            relevant_species.append(
                MixtureSpecies(
                    species=relevant_spec,
                    mole_fraction=spec[1],
                    state=state,
                )
            )

        temp_df = pd.DataFrame.from_records(
            data=[spec.model_dump() for spec in relevant_species]
        )
        species_df = pd.json_normalize(temp_df.species)
        temp_df.drop(columns=["species", "state"], inplace=True)
        df = pd.concat([temp_df, species_df], axis=1)
        df["mass_fraction"] = df.mass / df.mass.sum()
        df["mass_ovr_mole_weight"] = df.mass_fraction / df.molecular_weight
        super().__init__(species=species, state=state, df=df)

    @computed_field
    @property
    def molecular_weight(self) -> np.float64:
        return self.df.mole_fraction @ self.df.molecular_weight

    @computed_field
    @property
    def density(self) -> np.float64:
        return (self.state.pressure) / (
            R_U * (self.state.temperature) * self.df.mass_ovr_mole_weight.sum()
        )

    @computed_field
    @property
    def specific_heat_capacity(self) -> np.float64:
        return self.df.mass_fraction @ self.df.heat_capacity

    @computed_field
    @property
    def viscosity(self) -> np.float64:
        viscosity = 0
        for i in range(len(self.df)):
            series_i = self.df.iloc[i]
            num = series_i.mole_fraction * series_i.mu
            theta_num = (
                1
                + np.sqrt(series_i.mu / self.df.mu)
                * (self.df.molecular_weight / series_i.molecular_weight) ** 0.25
            ) ** 2
            theta_den = np.sqrt(
                (8 * (1 + (series_i.molecular_weight / self.df.molecular_weight)))
            )
            theta = theta_num / theta_den
            den = self.df.mole_fraction @ theta
            viscosity += num / den
        return viscosity

    @computed_field
    @property
    def conductivity(self) -> np.float64:
        conductivity = 0
        for i in range(len(self.df)):
            series_i = self.df.iloc[i]
            num = series_i.mole_fraction * series_i.k
            theta_num = (
                1
                + np.sqrt(series_i.mu / self.df.mu)
                * (self.df.molecular_weight / series_i.molecular_weight) ** 0.25
            ) ** 2
            theta_den = np.sqrt(
                (8 * (1 + (series_i.molecular_weight / self.df.molecular_weight)))
            )
            theta = theta_num / theta_den
            den = self.df.mole_fraction @ theta
            conductivity += num / den
        return conductivity

    @computed_field
    @property
    def adiabatic_index(self) -> np.float64:
        return self.specific_heat_capacity / (
            self.specific_heat_capacity - (R_U / self.molecular_weight)
        )

    @computed_field
    @property
    def speed_of_sound(self) -> np.float64:
        return np.sqrt(
            (
                self.adiabatic_index
                * (R_U / self.molecular_weight)
                * (self.state.temperature)
            )
        )



import argparse

import pandas as pd

from mixprops.mixture import Mixture
from mixprops.settings import get_settings
from mixprops.state import State

settings = get_settings()


def cli():
    parser = argparse.ArgumentParser(description="Mixture properties calculator.")
    parser.add_argument("--units", type=str, required=False, help="System units")
    parser.add_argument(
        "--pressure-unit", type=str, required=False, help="Pressure unit"
    )
    parser.add_argument(
        "--state",
        type=str,
        required=True,
        help="Mixture state",
    )
    parser.add_argument("--species", type=str, required=True, help="Species")
    parser.add_argument(
        "--composition",
        type=str,
        required=True,
        help="Molar composition as fractions.",
    )
    args = parser.parse_args()

    if args.units:
        settings.units = args.units
    if args.pressure_unit:
        settings.pressure_unit = args.pressure_unit

    state = args.state.split(",")

    state = State(
        settings=settings,
        pressure=float(state[0]),
        temperature=float(state[1]),
    )

    species = args.species.split(",")
    composition = [float(mf) for mf in args.composition.split(",")]

    if len(species) != len(composition):
        raise ValueError("Species and fractions should be of the same length")

    mixture = Mixture(species=list(zip(species, composition)), state=state)

    df = pd.DataFrame.from_dict([mixture.model_dump()])

    print("")
    print(df.T.rename(columns={0: "Value"}))
    print("")



from typing import Optional

from pydantic import (BaseModel, ConfigDict, Field, FieldValidationInfo,
                      field_validator)

from mixprops.constants import BAR_TO_PA, KELVIN_ADD, KPA_TO_PA, MPA_TO_PA
from mixprops.settings import Settings, get_settings
from mixprops.units import PressureUnit, UnitSystem


class State(BaseModel):
    settings: Optional[Settings] = Field(default_factory=get_settings)
    pressure: float
    temperature: float
    model_config = ConfigDict(use_enum_values=True)

    @field_validator("pressure")
    def validate_pressure(cls, value: float, info: FieldValidationInfo) -> float:
        pressure_unit: PressureUnit = info.data["settings"].pressure_unit
        if pressure_unit == PressureUnit.KILOPASCAL:
            value *= KPA_TO_PA
        elif pressure_unit == PressureUnit.MEGAPASCAL:
            value *= MPA_TO_PA
        elif pressure_unit == PressureUnit.BAR:
            value *= BAR_TO_PA
        return value

    @field_validator("temperature")
    def validate_temperature(cls, value: float, info: FieldValidationInfo) -> float:
        units = info.data["settings"].units
        if units == UnitSystem.SI_WITH_CELISUS:
            value += KELVIN_ADD
        elif units == UnitSystem.IMPERIAL:
            value = (value - 32) * (5 / 9) + KELVIN_ADD
        return value



import math
from typing import Optional

from pydantic import BaseModel, computed_field

from mixprops.constants import MU_CONSTANT, R_U, A, B, C, D, E, F
from mixprops.state import State


class Species(BaseModel):
    name: str
    species: str
    sigma: float
    eta: float
    molecular_weight: float
    a1: float
    a2: float
    a3: float
    a4: float
    a5: float
    a6: float
    a7: float
    a8: Optional[float]
    a9: Optional[float]


class MixtureSpecies(BaseModel):
    state: State
    species: Species
    mole_fraction: float

    @computed_field
    @property
    def mass(self) -> float:
        return self.mole_fraction * self.species.molecular_weight

    @computed_field
    @property
    def R(self) -> float:
        return R_U / self.species.molecular_weight

    @computed_field
    @property
    def heat_capacity(self) -> float:
        return self.R * (
            (self.species.a1 * self.state.temperature**-2)
            + (self.species.a2 * self.state.temperature**-1)
            + (self.species.a3)
            + (self.species.a4 * self.state.temperature)
            + (self.species.a5 * self.state.temperature**2)
            + (self.species.a6 * self.state.temperature**3)
            + (self.species.a7 * self.state.temperature**4)
        )

    @computed_field
    @property
    def t_star(self) -> float:
        return self.state.temperature / self.species.eta

    @computed_field
    @property
    def omega(self) -> float:
        return (
            A * (self.t_star**-B)
            + C * (math.exp(-D * self.t_star))
            + E * (math.exp(-F * self.t_star))
        )

    @computed_field
    @property
    def mu(self) -> float:
        return (
            MU_CONSTANT
            * math.sqrt(self.species.molecular_weight * self.state.temperature)
        ) / (self.species.sigma**2 * self.omega)

    @computed_field
    @property
    def k(self) -> float:
        return (
            (15 / 4)
            * (R_U / self.species.molecular_weight)
            * self.mu
            * (
                (4 / 15) * ((self.heat_capacity * self.species.molecular_weight) / R_U)
                + 1 / 3
            )
        )
